/**
 * @file main.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ----------------------------------------------------------------------------
 * (c) Florian Schmitt / 2012-2013
 * License GPLv3+: GNU GPL Version 3 
 * or higher <http://gnu.org/licenses/gpl.html>
 * This code comes with ABSOLUTELY NO WARRANTY. This is free software, and 
 * you are welcome to redistribute it under certain conditions.
 * ----------------------------------------------------------------------------
 * A simple program to set a channel's value on a lindmxd 
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include "config.h"


#define BUFLEN 1024
const char *program_name;
char buffer[BUFLEN];



/**
 * kill_lineend()
 * Kills the line end. The incoming string format does not matter: 
 * '\r', '\n', '\r\n' and '\n\r' is accepted.
 * @param str    String to modify
 * @param buflen buffer length, maximum place to go to
 * @return string length
**/
int kill_lineend(char *str, int buflen)
{
	int n = 0;
	/* look for the line/string end */
	while( n<buflen && (str[n] != '\r') && (str[n] != '\n') && (str[n] != '\0') ) ++n;
	/* place a \0 at the end */
	str[n] = '\0';
	return (n-1);
}




/**
 * print_usage()
 * Prints the program usage
 * @param stream     IOSTREAM to print to
 * @param exit_code  The code to give to exit() after printing
**/
void print_usage(FILE* stream, int exit_code)
{
	fprintf(stream, "Usage:  %s [options] channel value\n", program_name);
	fprintf(stream, 
			"Channel: The channels number (from 1) or \'all\'\n"
			"Value: Brightness (0-255)\n"
			"Options:\n"
			"  -h  --help           Display this usage information\n"
			"  -v  --version        Display version information\n"
			"  -p  --port n         Set the port to n\n"
			"  -i  --input <file>   Set an input file\n"
			"  -a  --additive       Add to current value\n"
			"  -s  --server <addr>  Set the host to connect to\n"
			"Value can be \'all\' to set all channels.\n"
			"Using an input file, channel and value are not needed.\n"
			"An input file has to contain one channel\'s value per line.\n");
	exit(exit_code);
}



/**
 * print_version()
 * Prints the program version
**/
void print_version()
{
	printf("%s: Version %d.%d.%d-%d\n", program_name, LINDMX_MAJOR_VERSION, 
	                                                   LINDMX_MINOR_VERSION, 
	                                                   LINDMX_PATCH_VERSION, 
	                                                   LINDMX_BUILD_NUMBER);
	printf("Copyright © 2013 Florian Schmitt\n"
	       "License GPLv3+: GNU GPL Version 3 or higher <http://gnu.org/licenses/gpl.html>\n"
	       "This program comes with ABSOLUTELY NO WARRANTY.\n"
	       "This is free software, and you are welcome to \n"
	       "redistribute it under certain conditions.\n\n"
	       "This software is written by Florian Schmitt <florian.schmitt@qikgrp.com>\n");
	exit(EXIT_SUCCESS);
}




/**
 * get_channel()
 * Gets the last value set for a channel
 * @param channel DMX-Channel number
 * @param sock    Socket connection
 * @return the channels value
**/
int get_channel(int channel, int sock)
{
	/* send data */
	sprintf(buffer, "get %d\r\n", channel);
	if( write(sock, buffer, strlen(buffer)) <= 0 ) 
	{
		fprintf(stderr, "Failed to send request.\n");
		exit(EXIT_FAILURE);
	}

	/* receive answer */
	if( read(sock, buffer, BUFLEN) <= 0 ) 
	{
		fprintf(stderr, "Failed to receive answer.\n");
		exit(EXIT_FAILURE);
	}

	/* return as int */
	kill_lineend(buffer, BUFLEN);
	return atoi(buffer);
}



/**
 * set_channel()
 * Sets a given channel on an active socket connection
 * @param channel     DMX-Channel number
 * @param value       Value to set to
 * @param sock        Socket connection
 * @param additive    if 1, the value will be added
**/
void set_channel(int channel, int value, int sock, int additive)
{
	if(additive) 
	{
		value += get_channel(channel, sock);
		if(value < 0) value = 0;
		else if(value > 255) value = 255;
	}
	/* send data */
	sprintf(buffer, "set %d %d\r\n", channel, value);
	if( write(sock, buffer, strlen(buffer)) <= 0 ) 
	{
		fprintf(stderr, "Failed to send request.\n");
		exit(EXIT_FAILURE);
	}

	/* receive answer */
	if( read(sock, buffer, BUFLEN) <= 0 ) 
	{
		fprintf(stderr, "Failed to receive answer.\n");
		exit(EXIT_FAILURE);
	}

	/* check answer */
	kill_lineend(buffer, BUFLEN);
	if( strcmp(buffer, "ok") != 0 )
	{
		fprintf(stderr, "Command failure: %s\n", buffer);
		exit(EXIT_FAILURE);
	}
}




/**
 * set_all_channels()
 * Sets all channels on an active socket connection
 * @param value   Value to set to
 * @param sock    Socket connection
**/
void set_all_channels(int value, int sock)
{
	/* send data */
	sprintf(buffer, "set * %d\r\n", value);
	if( write(sock, buffer, strlen(buffer)) <= 0 ) 
	{
		fprintf(stderr, "Failed to send request.\n");
		exit(EXIT_FAILURE);
	}

	/* receive answer */
	if( read(sock, buffer, BUFLEN) <= 0 ) 
	{
		fprintf(stderr, "Failed to receive answer.\n");
		exit(EXIT_FAILURE);
	}

	/* check answer */
	kill_lineend(buffer, BUFLEN);
	if( strcmp(buffer, "ok") != 0 )
	{
		fprintf(stderr, "Command failure: %s\n", buffer);
		exit(EXIT_FAILURE);
	}
}



/**
 * main()
 * Main procedure
**/
int main(int argc, char* argv[])
{
	int next_option, value, port, additive;
	const char *channel;
	const char *input;
	const char *remote = "localhost";

	/* options */
	const char* const short_options = "hvs:i:ap:";
	const struct option long_options[] = {
		{ "help",       0, NULL, 'h' },
		{ "version",    0, NULL, 'v' },
		{ "server",     1, NULL, 's' },
		{ "input",      1, NULL, 'i' },
		{ "additive",   0, NULL, 'a' },
		{ "port",       1, NULL, 'p' },
		{ NULL,       0, NULL,  0  }
	};

	/* default values */
	program_name = argv[0];
	additive = 0;
	channel  = 0;
	value    = 0;
	input    = 0;
	port     = 6767;

	/* get the options */
	do {
		next_option = getopt_long(argc, argv, short_options, long_options, NULL);
		switch(next_option)
		{
			case 'h': 
				print_usage(stdout, EXIT_SUCCESS);
				break;
			case 'v': 
				print_version();
				break;
			case 'a':
				additive = 1;
				break;
			case 'p': 
				port = atoi(optarg);
				break;
			case 's': 
				remote = optarg;
				break;
			case 'i':
				input = optarg;
				break;
			case '?': 
				print_usage(stderr, EXIT_FAILURE);
			case -1: 
				break;
			default:
				abort();
		}
	} while(next_option != -1);

	/* main parameters */
	if(input == 0)
	{
		if(argc <= optind+1) print_usage(stderr, EXIT_FAILURE);
		channel  = argv[optind];
		value    = atoi(argv[optind+1]);
	}

	/* client connection */
	int sock;
	struct hostent *host;
	struct sockaddr_in addr;

	if ( (host = gethostbyname(remote)) == NULL) 
	{
		fprintf(stderr, "Host '%s' can not be resolved\n", remote);
		exit(EXIT_FAILURE);
	}
	
	sock = socket(AF_INET, SOCK_STREAM, 0);
	addr.sin_family      = AF_INET;
	//addr.sin_addr.s_addr = inet_addr(remote);
	addr.sin_addr.s_addr = *((unsigned long *) host->h_addr_list[0]);
	addr.sin_port        = htons(port);
	
	/* try to connect */
	if( connect(sock, (struct sockaddr *)&addr, sizeof(addr)) != 0 )
	{
		fprintf(stderr, "Failed to connect to %s:%d\n", remote, port);
		exit(EXIT_FAILURE);
	}

	/* send single data */
	if(input == 0)
	{
		if(strcmp(channel, "all")==0) set_all_channels(value, sock);
		else set_channel(atoi(channel), value, sock, additive);
	}
	/* send data from file */
	else
	{
		/* open file */
		FILE *rf = fopen(input, "r");
		/* check file */
		if(rf == NULL) 
		{
			fprintf(stderr, "Failed to open file \'%s\'\n", input);
			exit(EXIT_FAILURE);
		}
		/* read the hole file */
		int n=1;
		while(fgets(buffer, sizeof(buffer), rf))
			set_channel(n++, atoi(buffer), sock, additive);
		/* close file */
		fclose(rf);
	}

	/* close connection */
	close(sock);

 	return 0;
}
